import type { FMember, FPMember } from './types';
export declare const PROMISE_FUNCTIONS: FPMember[];
export declare const CALLBACK_FUNCTIONS: FMember[];
export declare const SYNC_FUNCTIONS: FMember[];
//# sourceMappingURL=constants.d.ts.map